/*
 * GL-Zoomquilt. See http://zoomquilt.nikkki.net/
 */

#include <stdlib.h>
#include <math.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include <gtk/gtkgl.h>
#include <GL/gl.h>

typedef struct _QuiltWindow   QuiltWindow;

#define N_QUILT_IMAGES 46

struct _QuiltWindow
{
  GtkWidget    *window;
  GtkWidget    *drawing_area;
  GTimer       *timer;
  guint         timeout_id;
  gboolean      visibility;

  gdouble       fps;
  gdouble       time;

  gdouble       speed;
  gdouble       zoom;

  GLuint        textures[N_QUILT_IMAGES];
};


static gboolean
quilt_frame_next (QuiltWindow *hwin)
{
  gdouble time, delta_time;

  time = g_timer_elapsed (hwin->timer, NULL);

  delta_time = time - hwin->time;
  hwin->time = time;

  hwin->zoom += hwin->speed * delta_time;

  if (hwin->visibility)
    gtk_widget_queue_draw (hwin->drawing_area);

  return TRUE;
}


/* draw a frame */
static gboolean
quilt_frame_draw (GtkWidget      *widget,
                  GdkEventExpose *event,
                  gpointer        data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);
  gint j, j0;
  gdouble tf;
  GLfloat color[4] = {1.0, 1.0, 1.0, 1.0};
  QuiltWindow  *hwin = (QuiltWindow *) data;

  /*** OpenGL BEGIN ***/
  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return FALSE;

  glClear (GL_COLOR_BUFFER_BIT);
  glEnable (GL_TEXTURE_2D);

  glMaterialfv (GL_FRONT, GL_AMBIENT_AND_DIFFUSE, color);
  glNormal3f (0, 0, 1.0);

  while (hwin->zoom < 0)
    hwin->zoom += N_QUILT_IMAGES;

  while (hwin->zoom > N_QUILT_IMAGES)
    hwin->zoom -= N_QUILT_IMAGES;

  j0 = floor (hwin->zoom);

  glPushMatrix ();
  tf = pow (2.0, hwin->zoom - j0);
  glScalef (tf, tf, 1.0);

  for (j = 0; j < 3; j++)
    {
      glBindTexture (GL_TEXTURE_2D, hwin->textures[(j0 + j) % N_QUILT_IMAGES]);

      glBegin(GL_POLYGON);
      glTexCoord2f (0.0, 1.0);
      glVertex2f   (-1.0, -1.0);
      glTexCoord2f (1.0, 1.0);
      glVertex2f   (1.0, -1.0);
      glTexCoord2f (1.0, 0.0);
      glVertex2f   (1.0, 1.0);
      glTexCoord2f (0.0, 0.0);
      glVertex2f   (-1.0, 1.0);
      glEnd ();

      glScalef (0.5, 0.5, 1.0);
    }
  glPopMatrix ();

  if (gdk_gl_drawable_is_double_buffered (gldrawable))
    gdk_gl_drawable_swap_buffers (gldrawable);
  else
    glFlush ();

  gdk_gl_drawable_gl_end (gldrawable);
  /*** OpenGL END ***/

  return TRUE;
}


/* new window size or exposure */
static gboolean
quilt_configure_event (GtkWidget         *widget,
                       GdkEventConfigure *event,
                       gpointer           data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  GLfloat h = (GLfloat) (widget->allocation.height) / (GLfloat) (widget->allocation.width) / 0.75;

  /*** OpenGL BEGIN ***/
  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return FALSE;

  glViewport (0, 0, widget->allocation.width, widget->allocation.height);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  if (h < 1.0)
    glOrtho (-1.0, 1.0, -1.0*h, 1.0*h, -1.0, 1.0);
  else
    glOrtho (-1.0/h, 1.0/h, -1.0, 1.0, -1.0, 1.0);

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();

  gdk_gl_drawable_gl_end (gldrawable);
  /*** OpenGL END ***/

  return TRUE;
}


static gboolean
quilt_visibility_update (GtkWidget   *widget,
                         GdkEventAny *event,
                         gpointer     data)
{
  QuiltWindow *hwin = (QuiltWindow *) data;

  switch (event->type)
    {
      case GDK_MAP:
          hwin->visibility = TRUE;
          break;
      case GDK_UNMAP:
          hwin->visibility = FALSE;
          break;
      case GDK_VISIBILITY_NOTIFY:
          hwin->visibility = ((GdkEventVisibility *) event)->state != GDK_VISIBILITY_FULLY_OBSCURED;
          break;
      default:
          g_printerr ("Unknown event in quilt_visibility_update!\n");
          hwin->visibility = TRUE;
          break;
    }

  return TRUE;
}


/* change view angle, exit upon ESC */
static gboolean
quilt_key_press_event (GtkWidget   *widget,
                       GdkEventKey *event,
                       gpointer     data)
{
  static gboolean fullscreen = FALSE;
  static gdouble old_speed = 0.0;
  QuiltWindow *hwin = (QuiltWindow *) data;

  switch (event->keyval)
    {
    case GDK_F11:
    case GDK_F:
    case GDK_f:
      if (fullscreen)
        gtk_window_unfullscreen (GTK_WINDOW (hwin->window));
      else
        gtk_window_fullscreen (GTK_WINDOW (hwin->window));
      fullscreen = !fullscreen;
      break;
    case GDK_Escape:
    case GDK_Q:
    case GDK_q:
      gtk_main_quit ();
      break;
    case GDK_Up:
      hwin->zoom += 0.05;
      if (fabs (hwin->speed) > 0.00001)
        {
          old_speed = hwin->speed;
          hwin->speed = 0.0;
        }
      break;
    case GDK_Down:
      hwin->zoom -= 0.05;
      if (fabs (hwin->speed) > 0.00001)
        {
          old_speed = hwin->speed;
          hwin->speed = 0.0;
        }
      break;
    case GDK_Left:
      hwin->speed += 0.05;
      break;
    case GDK_Right:
      hwin->speed -= 0.05;
      break;
    case GDK_space:
    case GDK_P:
    case GDK_p:
      if (fabs (hwin->speed) > 0.00001)
        {
          old_speed = hwin->speed;
          hwin->speed = 0.0;
        }
      else
        {
          hwin->speed = old_speed;
          old_speed = 0.0;
        }
      break;
    default:
      return FALSE;
    }

  return TRUE;
}


static void
quilt_init_gl (GtkWidget *widget,
               gpointer   data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);
  QuiltWindow *hwin = (QuiltWindow *) data;
  gint i;

  static GLfloat pos[4] = {5.0, 5.0, 10.0, 0.0};

  /*** OpenGL BEGIN ***/
  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return;

  glLightfv (GL_LIGHT0, GL_POSITION, pos);
  glEnable (GL_LIGHTING);
  glEnable (GL_LIGHT0);

  glEnable (GL_NORMALIZE);

  glGenTextures (N_QUILT_IMAGES, hwin->textures);

  for (i = 0; i < N_QUILT_IMAGES; i++)
    {
      GdkPixbuf *pixbuf, *pixbuf2;
      guchar *data, *filename;

      filename = g_strdup_printf ("merged/zq-%02d.jpg", N_QUILT_IMAGES - i);

      pixbuf2 = gdk_pixbuf_new_from_file (filename, NULL);
      pixbuf = gdk_pixbuf_scale_simple (pixbuf2, 1024, 1024,
                                        GDK_INTERP_BILINEAR);

      g_object_unref (pixbuf2);

      g_printerr ("%s: %d, %d - %d = %d\n", filename,
                  gdk_pixbuf_get_width (pixbuf),
                  gdk_pixbuf_get_height (pixbuf),
                  gdk_pixbuf_get_width (pixbuf) *
                  gdk_pixbuf_get_n_channels (pixbuf),
                  gdk_pixbuf_get_rowstride (pixbuf));
      data = gdk_pixbuf_get_pixels (pixbuf);

      glBindTexture (GL_TEXTURE_2D, hwin->textures[i]);

      glPixelStorei (GL_UNPACK_ALIGNMENT, 1);
      glTexImage2D (GL_TEXTURE_2D, 0, 3,
                    gdk_pixbuf_get_width (pixbuf),
                    gdk_pixbuf_get_height (pixbuf),
                    0, GL_RGB,
                    GL_UNSIGNED_BYTE,
                    gdk_pixbuf_get_pixels (pixbuf));
      glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
      glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
      glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
      glShadeModel (GL_FLAT);

      g_object_unref (pixbuf);
      g_free (filename);
    }


#if 0
  g_printerr ("\n");
  g_printerr ("GL_RENDERER   = %s\n", (char *) glGetString (GL_RENDERER));
  g_printerr ("GL_VERSION    = %s\n", (char *) glGetString (GL_VERSION));
  g_printerr ("GL_VENDOR     = %s\n", (char *) glGetString (GL_VENDOR));
  g_printerr ("GL_EXTENSIONS = %s\n", (char *) glGetString (GL_EXTENSIONS));
  g_printerr ("\n");
#endif

  gdk_gl_drawable_gl_end (gldrawable);
  /*** OpenGL END ***/

  /* create timer */
  if (hwin->timer == NULL)
    hwin->timer = g_timer_new ();

  g_timer_start (hwin->timer);
}


QuiltWindow *
quilt_setup_ui ()
{
  GdkGLConfig *glconfig;
  GtkWidget *vbox;
  QuiltWindow *hwin;

  hwin = g_new0 (QuiltWindow, 1);

  hwin->fps = 50;
  hwin->speed = -0.15;
  hwin->zoom = g_random_double_range (0.0, N_QUILT_IMAGES);

  /*
   * Configure OpenGL-capable visual.
   */

  /* Try double-buffered visual */
  glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGB    |
                                        GDK_GL_MODE_DOUBLE);
  if (glconfig == NULL)
    {
      g_printerr ("*** Cannot find the double-buffered visual.\n");
      g_printerr ("*** Trying single-buffered visual.\n");

      /* Try single-buffered visual */
      glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGB);
      if (glconfig == NULL)
        {
          g_printerr ("*** No appropriate OpenGL-capable visual found.\n");
          exit (1);
        }
    }

  /*
   * Top-level window.
   */

  hwin->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (hwin->window), "Zoomquilt");

  /* Get automatically redrawn if any of their children changed allocation. */
  gtk_container_set_reallocate_redraws (GTK_CONTAINER (hwin->window), TRUE);

  g_signal_connect (G_OBJECT (hwin->window), "delete_event",
                    G_CALLBACK (gtk_main_quit), NULL);

  /*
   * VBox.
   */

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (hwin->window), vbox);
  gtk_widget_show (vbox);

  /*
   * Drawing area for drawing OpenGL scene.
   */

  hwin->drawing_area = gtk_drawing_area_new ();
  gtk_widget_set_size_request (hwin->drawing_area, 800, 600);

  /* Set OpenGL-capability to the widget. */
  gtk_widget_set_gl_capability (hwin->drawing_area,
                                glconfig,
                                NULL,
                                TRUE,
                                GDK_GL_RGBA_TYPE);

  gtk_widget_add_events (hwin->drawing_area,
                         GDK_VISIBILITY_NOTIFY_MASK);

  g_signal_connect_after (G_OBJECT (hwin->drawing_area), "realize",
                          G_CALLBACK (quilt_init_gl), hwin);
  g_signal_connect (G_OBJECT (hwin->drawing_area), "configure_event",
                    G_CALLBACK (quilt_configure_event), NULL);
  g_signal_connect (G_OBJECT (hwin->drawing_area), "expose_event",
                    G_CALLBACK (quilt_frame_draw), hwin);

  g_signal_connect (G_OBJECT (hwin->drawing_area), "map_event",
                    G_CALLBACK (quilt_visibility_update), hwin);
  g_signal_connect (G_OBJECT (hwin->drawing_area), "unmap_event",
                    G_CALLBACK (quilt_visibility_update), hwin);
  g_signal_connect (G_OBJECT (hwin->drawing_area), "visibility_notify_event",
                    G_CALLBACK (quilt_visibility_update), hwin);

  g_signal_connect (G_OBJECT (hwin->window), "key_press_event",
                    G_CALLBACK (quilt_key_press_event), hwin);

  gtk_box_pack_start (GTK_BOX (vbox), hwin->drawing_area, TRUE, TRUE, 0);

  gtk_widget_show (hwin->drawing_area);

  /*
   * Show window.
   */

  gtk_widget_show (hwin->window);

  /*
   * Timeout function to maintain framerate
   */
  hwin->timeout_id = g_timeout_add_full (GDK_PRIORITY_REDRAW,
                                         1000 / hwin->fps,
                                         (GSourceFunc) quilt_frame_next,
                                         hwin,
                                         NULL);
  return hwin;
}


int
main (int   argc,
      char *argv[])
{
  QuiltWindow *hwin;

  gtk_init (&argc, &argv);
  gtk_gl_init (&argc, &argv);

  hwin = quilt_setup_ui ();

  gtk_main ();

  return 0;
}

